@@ -311,10 +311,7 @@ class Debusine:
         """Return URL to upload a file in the artifact."""
         quoted_path = urllib.parse.quote(path_in_artifact)
 
-        return (
-            f"{self.base_api_url}/{self.API_VERSION}/artifact/"
-            f"{artifact_id}/files/{quoted_path}/"
-        )
+        return f"{self.api_url}/artifact/{artifact_id}/files/{quoted_path}/"
 
     def upload_files(
         self,
