@@ -1,4 +1,4 @@
-# Copyright 2019, 2021-2024 The Debusine Developers
+# Copyright © The Debusine Developers
 # See the AUTHORS file at the top-level directory of this distribution
 #
 # This file is part of Debusine. It is subject to the license terms
@@ -12,19 +12,26 @@
 from datetime import datetime, timedelta
 from functools import partial
 from pathlib import Path
-from typing import Any, Optional, TYPE_CHECKING
+from typing import Any, Generic, Optional, TYPE_CHECKING, TypeVar
 
 from django.conf import settings
-from django.core.exceptions import ValidationError
+from django.core.exceptions import PermissionDenied, ValidationError
 from django.db import models, transaction
-from django.db.models import F, Q, QuerySet, UniqueConstraint
+from django.db.models import Exists, F, OuterRef, Q, QuerySet, UniqueConstraint
 from django.db.models.functions import Coalesce
 from django.urls import reverse
 
 from debusine.artifacts import LocalArtifact
 from debusine.artifacts.models import ArtifactCategory, ArtifactData
+from debusine.db.context import context
 from debusine.db.models.files import File
+from debusine.db.models.permissions import (
+    PermissionUser,
+    permission_check,
+    permission_filter,
+)
 from debusine.db.models.workspaces import Workspace
+from debusine.utils.typing_utils import copy_signature_from
 
 if TYPE_CHECKING:
     from django_stubs_ext.db.models import TypedModelMeta
@@ -33,6 +40,8 @@ if TYPE_CHECKING:
 else:
     TypedModelMeta = object
 
+A = TypeVar("A")
+
 ARTIFACT_CATEGORY_ICON_NAMES = {
     ArtifactCategory.AUTOPKGTEST: "folder",
     ArtifactCategory.BINARY_PACKAGE: "folder",
@@ -42,7 +51,6 @@ ARTIFACT_CATEGORY_ICON_NAMES = {
     ArtifactCategory.LINTIAN: "folder",
     ArtifactCategory.PACKAGE_BUILD_LOG: "journal-text",
     ArtifactCategory.SIGNING_INPUT: "folder",
-    ArtifactCategory.SIGNING_KEY: "folder",
     ArtifactCategory.SIGNING_OUTPUT: "folder",
     ArtifactCategory.SOURCE_PACKAGE: "folder",
     ArtifactCategory.SYSTEM_IMAGE: "folder",
@@ -60,7 +68,6 @@ ARTIFACT_CATEGORY_SHORT_NAMES = {
     ArtifactCategory.LINTIAN: "lintian report",
     ArtifactCategory.PACKAGE_BUILD_LOG: "build log",
     ArtifactCategory.SIGNING_INPUT: "signing input",
-    ArtifactCategory.SIGNING_KEY: "signing key",
     ArtifactCategory.SIGNING_OUTPUT: "signing output",
     ArtifactCategory.SOURCE_PACKAGE: "source package",
     ArtifactCategory.SYSTEM_IMAGE: "system image",
@@ -70,9 +77,44 @@ ARTIFACT_CATEGORY_SHORT_NAMES = {
 }
 
 
+class ArtifactQuerySet(QuerySet["Artifact", A], Generic[A]):
+    """Custom QuerySet for Artifact."""
+
+    def in_current_scope(self) -> "ArtifactQuerySet[A]":
+        """Filter to artifacts in the current scope."""
+        return self.filter(workspace__scope=context.require_scope())
+
+    def in_current_workspace(self) -> "ArtifactQuerySet[A]":
+        """Filter to artifacts in the current workspace."""
+        return self.filter(workspace=context.require_workspace())
+
+    @permission_filter
+    def can_display(self, user: PermissionUser) -> "ArtifactQuerySet[A]":
+        """Keep only Artifacts that can be displayed."""
+        # Delegate to workspace can_display check
+        return self.filter(workspace__in=Workspace.objects.can_display(user))
+
+    # mypy_django_plugin generates an
+    # Artifact@AnnotatedWith[TypedDict({"complete": Any})] model subclass
+    # for this, but we don't have a way to refer to it here.
+    def annotate_complete(self) -> "ArtifactQuerySet[Any]":
+        """Annotate artifacts with whether all their files are complete."""
+        return self.annotate(
+            complete=~Exists(
+                FileInArtifact.objects.filter(
+                    artifact=OuterRef("pk"), complete=False
+                )
+            )
+        )
+
+
 class ArtifactManager(models.Manager["Artifact"]):
     """Manager for the Artifact model."""
 
+    def get_queryset(self) -> ArtifactQuerySet[Any]:
+        """Use the custom QuerySet."""
+        return ArtifactQuerySet(self.model, using=self._db)
+
     @classmethod
     def create_from_local_artifact(
         cls,
@@ -82,8 +124,6 @@ class ArtifactManager(models.Manager["Artifact"]):
         created_by_work_request: Optional["WorkRequest"] = None,
     ) -> "Artifact":
         """Return a new Artifact based on a :class:`LocalArtifact`."""
-        file_store = workspace.default_file_store.get_backend_object()
-
         artifact = Artifact.objects.create(
             category=local_artifact.category,
             workspace=workspace,
@@ -93,7 +133,8 @@ class ArtifactManager(models.Manager["Artifact"]):
 
         for artifact_path, local_path in local_artifact.files.items():
             file = File.from_local_path(local_path)
-            file_store.add_file(local_path, fileobj=file)
+            file_backend = workspace.scope.upload_file_backend(file)
+            file_backend.add_file(local_path, fileobj=file)
             FileInArtifact.objects.create(
                 artifact=artifact, path=artifact_path, file=file, complete=True
             )
@@ -194,9 +235,9 @@ class Artifact(models.Model):
         "Artifact", blank=True, null=True, on_delete=models.SET_NULL
     )
 
-    objects = ArtifactManager()
+    objects = ArtifactManager.from_queryset(ArtifactQuerySet)()
 
-    def clean(self):
+    def clean(self) -> None:
         """
         Ensure that data is valid for this artifact category.
 
@@ -219,10 +260,35 @@ class Artifact(models.Model):
                 {"data": f"invalid artifact data: {e}"}
             ) from e
 
+    @permission_check("{user} cannot display artifact {resource}")
+    def can_display(self, user: PermissionUser) -> bool:
+        """Check if the artifact can be displayed."""
+        return self.workspace.can_display(user)
+
     def create_data(self) -> ArtifactData:
         """Instantiate ArtifactData from data."""
         artifact_cls = LocalArtifact.class_from_category(self.category)
-        return artifact_cls.create_data(self.data)
+        artifact_data = artifact_cls.create_data(self.data)
+        assert isinstance(artifact_data, ArtifactData)
+        return artifact_data
+
+    @copy_signature_from(models.Model.save)
+    def save(self, **kwargs: Any) -> None:
+        """Wrap save with permission checks."""
+        from debusine.db.context import context
+
+        if self._state.adding:
+            # Create
+            if not self.workspace.can_create_artifacts(context.user):
+                raise PermissionDenied(
+                    f"{context.user} cannot create artifacts"
+                    f" in {self.workspace}"
+                )
+        else:
+            # Update
+            ...  # TODO: check for update permissions
+
+        return super().save(**kwargs)
 
     def get_label(self, data: ArtifactData | None = None) -> str:
         """
@@ -236,11 +302,12 @@ class Artifact(models.Model):
             return label
         return str(self.category)
 
-    def effective_expiration_delay(self):
+    def effective_expiration_delay(self) -> timedelta:
         """Return expiration_delay, inherited if None."""
         expiration_delay = self.expiration_delay
         if self.expiration_delay is None:  # inherit
             expiration_delay = self.workspace.default_expiration_delay
+        assert expiration_delay is not None
         return expiration_delay
 
     @property
@@ -266,7 +333,17 @@ class Artifact(models.Model):
 
     def get_absolute_url(self) -> str:
         """Return the canonical URL to display the artifact."""
-        return reverse("artifacts:detail", kwargs={"artifact_id": self.id})
+        return reverse(
+            "workspaces:artifacts:detail",
+            kwargs={"wname": self.workspace.name, "artifact_id": self.id},
+        )
+
+    def get_absolute_url_download(self) -> str:
+        """Return the canonical URL to download the artifact."""
+        return reverse(
+            "workspaces:artifacts:download",
+            kwargs={"wname": self.workspace.name, "artifact_id": self.id},
+        )
 
     def __str__(self) -> str:
         """Return basic information of Artifact."""
@@ -305,6 +382,39 @@ class FileInArtifact(models.Model):
             f"Path: {self.path} File: {self.file.id}"
         )
 
+    def get_absolute_url(self) -> str:
+        """Return an absolute URL to view this file."""
+        return reverse(
+            "workspaces:artifacts:file-detail",
+            kwargs={
+                "wname": self.artifact.workspace.name,
+                "artifact_id": self.artifact.id,
+                "path": self.path,
+            },
+        )
+
+    def get_absolute_url_raw(self) -> str:
+        """Return an absolute URL to view this file as raw."""
+        return reverse(
+            "workspaces:artifacts:file-raw",
+            kwargs={
+                "wname": self.artifact.workspace.name,
+                "artifact_id": self.artifact.id,
+                "path": self.path,
+            },
+        )
+
+    def get_absolute_url_download(self) -> str:
+        """Return an absolute URL to download this file."""
+        return reverse(
+            "workspaces:artifacts:file-download",
+            kwargs={
+                "wname": self.artifact.workspace.name,
+                "artifact_id": self.artifact.id,
+                "path": self.path,
+            },
+        )
+
 
 class FileUpload(models.Model):
     """File that is being/has been uploaded."""
@@ -356,7 +466,7 @@ class FileUpload(models.Model):
 
         return size
 
-    def delete(self, *args, **kwargs):
+    def delete(self, *args: Any, **kwargs: Any) -> tuple[int, dict[str, int]]:
         """Schedule deletion of the file in the store."""
         file_path = self.absolute_file_path()
 
@@ -389,6 +499,36 @@ class FileUpload(models.Model):
         return f"{self.id}"
 
 
+class ArtifactRelationQuerySet(QuerySet["ArtifactRelation", A], Generic[A]):
+    """Custom QuerySet for ArtifactRelation."""
+
+    def in_current_scope(self) -> "ArtifactRelationQuerySet[A]":
+        """Filter to artifact relations in the current scope."""
+        scope = context.require_scope()
+        return self.filter(
+            artifact__workspace__scope=scope, target__workspace__scope=scope
+        )
+
+    @permission_filter
+    def can_display(
+        self, user: PermissionUser
+    ) -> "ArtifactRelationQuerySet[A]":
+        """Keep only ArtifactRelations that can be displayed."""
+        # Delegate to workspace can_display check
+        workspaces = Workspace.objects.can_display(user)
+        return self.filter(
+            artifact__workspace__in=workspaces, target__workspace__in=workspaces
+        )
+
+
+class ArtifactRelationManager(models.Manager["ArtifactRelation"]):
+    """Manager for the ArtifactRelation model."""
+
+    def get_queryset(self) -> ArtifactRelationQuerySet[Any]:
+        """Use the custom QuerySet."""
+        return ArtifactRelationQuerySet(self.model, using=self._db)
+
+
 class ArtifactRelation(models.Model):
     """Model relations between artifacts."""
 
@@ -397,6 +537,8 @@ class ArtifactRelation(models.Model):
         RELATES_TO = "relates-to", "Relates to"
         BUILT_USING = "built-using", "Built using"
 
+    objects = ArtifactRelationManager.from_queryset(ArtifactRelationQuerySet)()
+
     artifact = models.ForeignKey(
         Artifact, on_delete=models.PROTECT, related_name="relations"
     )
@@ -416,3 +558,10 @@ class ArtifactRelation(models.Model):
                 name="%(app_label)s_%(class)s_unique_artifact_target_type",
             )
         ]
+
+    @permission_check("{user} cannot display artifact relation {resource}")
+    def can_display(self, user: PermissionUser) -> bool:
+        """Check if the artifact can be displayed."""
+        return self.artifact.workspace.can_display(
+            user
+        ) and self.target.workspace.can_display(user)
