@@ -1,4 +1,4 @@
-# Copyright 2019, 2021-2024 The Debusine Developers
+# Copyright © The Debusine Developers
 # See the AUTHORS file at the top-level directory of this distribution
 #
 # This file is part of Debusine. It is subject to the license terms
@@ -9,18 +9,18 @@
 
 """Unit tests for the worker models."""
 
+import datetime
 from typing import ClassVar
 
 from django.utils import timezone
 
-from debusine.db.models import Token, WorkRequest, Worker
+from debusine.db.models import Token, WorkRequest, Worker, WorkerPool
 from debusine.db.models.workers import WorkerManager
 from debusine.tasks.models import WorkerType
-from debusine.test import TestHelpersMixin
 from debusine.test.django import TestCase
 
 
-class WorkerManagerTests(TestHelpersMixin, TestCase):
+class WorkerManagerTests(TestCase):
     """Tests for the WorkerManager."""
 
     def test_connected(self) -> None:
@@ -38,6 +38,70 @@ class WorkerManagerTests(TestHelpersMixin, TestCase):
 
         self.assertQuerySetEqual(Worker.objects.connected(), [worker_connected])
 
+    def test_active(self) -> None:
+        """Test active() filter."""
+        worker_pool = self.playground.create_worker_pool()
+        workers: dict[str, Worker] = {}
+
+        workers["static_disconnected"] = self.playground.create_worker(
+            fqdn="disconnected.static.lan"
+        )
+
+        workers["static_connected_idle"] = self.playground.create_worker(
+            fqdn="idle.connected.static.lan"
+        )
+
+        workers["dynamic_terminated"] = self.playground.create_worker(
+            fqdn="disconnected.dynamic.lan", worker_pool=worker_pool
+        )
+        workers["dynamic_terminated"].instance_created_at = None
+        workers["dynamic_terminated"].save()
+
+        workers["dynamic_disconnected"] = self.playground.create_worker(
+            fqdn="disconnected.dynamic.lan", worker_pool=worker_pool
+        )
+
+        workers["dynamic_connected_idle"] = self.playground.create_worker(
+            fqdn="idle.connected.dynamic.lan", worker_pool=worker_pool
+        )
+        workers["dynamic_connected_idle"].connected_at = timezone.now()
+        workers["dynamic_connected_idle"].save()
+
+        workers["dynamic_connected_busy"] = self.playground.create_worker(
+            fqdn="busy.connected.dynamic.lan", worker_pool=worker_pool
+        )
+        workers["dynamic_connected_busy"].connected_at = timezone.now()
+        workers["dynamic_connected_busy"].save()
+        wr = self.playground.create_work_request()
+        wr.assign_worker(workers["dynamic_connected_busy"])
+        wr.mark_running()
+
+        workers["dynamic_connected_now_idle"] = self.playground.create_worker(
+            fqdn="now_idle.connected.dynamic.lan", worker_pool=worker_pool
+        )
+        workers["dynamic_connected_now_idle"].connected_at = timezone.now()
+        workers["dynamic_connected_now_idle"].save()
+        wr = self.playground.create_work_request()
+        wr.assign_worker(workers["dynamic_connected_now_idle"])
+        wr.mark_completed(WorkRequest.Results.SUCCESS)
+
+        for worker, included in (
+            ("static_disconnected", True),
+            ("static_connected_idle", True),
+            ("dynamic_terminated", False),
+            ("dynamic_disconnected", True),
+            ("dynamic_connected_idle", True),
+            ("dynamic_connected_busy", True),
+            ("dynamic_connected_now_idle", True),
+        ):
+            with self.subTest(worker=worker):
+                self.assertEqual(
+                    Worker.objects.active()
+                    .filter(pk=workers[worker].pk)
+                    .exists(),
+                    included,
+                )
+
     def test_waiting_for_work_request(self) -> None:
         """Test WorkerManager.waiting_for_work_request() return a Worker."""
         worker = Worker.objects.create_with_fqdn(
@@ -128,6 +192,32 @@ class WorkerManagerTests(TestHelpersMixin, TestCase):
             Worker.objects.waiting_for_work_request(), [worker_enabled]
         )
 
+    def test_waiting_for_work_request_static_first(self) -> None:
+        """Ensure waiting_for_work_request() returns static workers first."""
+        worker_static = Worker.objects.create_with_fqdn(
+            "worker-static", Token.objects.create(enabled=True)
+        )
+
+        worker_cloud = Worker.objects.create_with_fqdn(
+            "worker-cloud", Token.objects.create(enabled=True)
+        )
+        worker_cloud.worker_pool = WorkerPool.objects.create(
+            name="test",
+            provider_account=(
+                self.playground.create_cloud_provider_account_asset()
+            ),
+            registered_at=timezone.now(),
+        )
+        worker_cloud.save()
+
+        worker_cloud.mark_connected()
+        worker_static.mark_connected()
+
+        self.assertQuerySetEqual(
+            Worker.objects.waiting_for_work_request(),
+            [worker_static, worker_cloud],
+        )
+
     def test_create_with_fqdn_new_fqdn(self) -> None:
         """WorkerManager.create_with_fqdn() return a worker."""
         token = Token.objects.create()
@@ -159,6 +249,35 @@ class WorkerManagerTests(TestHelpersMixin, TestCase):
         self.assertEqual(worker.worker_type, WorkerType.EXTERNAL)
         self.assertIsNotNone(worker.pk)
 
+    def test_create_pool_members(self) -> None:
+        worker_pool = self.playground.create_worker_pool(name="pool")
+        # 1 existing worker in the middle of our scheme
+        Worker.objects.create_with_fqdn(
+            "pool-002", token=Token.objects.create()
+        )
+
+        before = timezone.now()
+        Worker.objects.create_pool_members(worker_pool, 3)
+        after = timezone.now()
+
+        self.assertQuerySetEqual(
+            Worker.objects.filter(
+                worker_pool=worker_pool,
+                activation_token__isnull=False,
+                activation_token__expire_at__gte=(
+                    before + datetime.timedelta(minutes=15)
+                ),
+                activation_token__expire_at__lte=(
+                    after + datetime.timedelta(minutes=15)
+                ),
+                activation_token__enabled=True,
+                token__isnull=True,
+            )
+            .order_by("name")
+            .values_list("name", flat=True),
+            ["pool-001", "pool-003", "pool-004"],
+        )
+
     def test_get_or_create_celery(self) -> None:
         """WorkerManager.get_or_create_celery returns a Celery worker."""
         worker = Worker.objects.get_or_create_celery()
@@ -210,8 +329,75 @@ class WorkerManagerTests(TestHelpersMixin, TestCase):
         """WorkerManager.get_worker_or_none() return None."""
         self.assertIsNone(Worker.objects.get_worker_or_none('does-not-exist'))
 
+    def test_with_idle_time_no_work_requests_not_connected(self) -> None:
+        """Test idle_time with no work requests assigned and never connected."""
+        worker = Worker.objects.create_with_fqdn(
+            'worker-a', Token.objects.create(enabled=True)
+        )
+        worker.registered_at = timezone.now() - datetime.timedelta(days=1)
+        worker.save()
 
-class WorkerTests(TestHelpersMixin, TestCase):
+        idle_time = (
+            Worker.objects.with_idle_time().get(pk=worker.pk)
+        ).idle_time
+        self.assertGreaterEqual(idle_time, 86400)
+        self.assertLessEqual(idle_time, 86400 + 3600)
+
+    def test_with_idle_time_no_work_requests_connected(self) -> None:
+        """Compute idle_time with no work requests assigned and connected."""
+        worker = Worker.objects.create_with_fqdn(
+            'worker-a', Token.objects.create(enabled=True)
+        )
+        worker.registered_at = timezone.now() - datetime.timedelta(days=2)
+        worker.instance_created_at = timezone.now() - datetime.timedelta(days=1)
+        worker.save()
+
+        idle_time = (
+            Worker.objects.with_idle_time().get(pk=worker.pk)
+        ).idle_time
+        self.assertGreaterEqual(idle_time, 86400)
+        self.assertLessEqual(idle_time, 86400 + 3600)
+
+    def test_with_idle_time_running_task(self) -> None:
+        """Compute idle_time with a running task."""
+        worker = Worker.objects.create_with_fqdn(
+            'worker-a', Token.objects.create(enabled=True)
+        )
+        worker.registered_at = timezone.now() - datetime.timedelta(days=2)
+        worker.instance_created_at = timezone.now() - datetime.timedelta(days=1)
+        worker.save()
+        work_request = self.playground.create_work_request(worker=worker)
+        work_request.mark_running()
+        worker.refresh_from_db()
+        self.assertTrue(worker.is_busy())
+
+        idle_time = (
+            Worker.objects.with_idle_time().get(pk=worker.pk)
+        ).idle_time
+        self.assertGreaterEqual(idle_time, 0)
+
+    def test_with_idle_time_completed_task(self) -> None:
+        """Compute idle_time with a running task."""
+        worker = Worker.objects.create_with_fqdn(
+            'worker-a', Token.objects.create(enabled=True)
+        )
+        worker.registered_at = timezone.now() - datetime.timedelta(days=2)
+        worker.instance_created_at = timezone.now() - datetime.timedelta(days=1)
+        worker.save()
+        work_request = self.playground.create_work_request(worker=worker)
+        work_request.mark_running()
+        work_request.mark_completed(WorkRequest.Results.SUCCESS)
+        work_request.completed_at = timezone.now() - datetime.timedelta(hours=6)
+        work_request.save()
+
+        idle_time = (
+            Worker.objects.with_idle_time().get(pk=worker.pk)
+        ).idle_time
+        self.assertGreaterEqual(idle_time, 6 * 3600)
+        self.assertLessEqual(idle_time, 7 * 3600)
+
+
+class WorkerTests(TestCase):
     """Tests for the Worker model."""
 
     worker: ClassVar[Worker]
@@ -293,7 +479,7 @@ class WorkerTests(TestHelpersMixin, TestCase):
         work_request.mark_aborted()
         self.assertFalse(worker.is_busy())
 
-    def test_running_work_requests(self):
+    def test_running_work_requests(self) -> None:
         """The running_work_requests return expected work requests."""
         worker_1 = self.playground.create_worker()
         worker_1.concurrency = 3
@@ -338,12 +524,10 @@ class WorkerTests(TestHelpersMixin, TestCase):
 
     def test_metadata_is_deep_copy(self) -> None:
         """Test metadata does a deep copy."""
-        self.worker.dynamic_metadata['schroots'] = ['buster', 'bullseye']
-        self.worker.metadata()['schroots'].append('bookworm')
+        self.worker.dynamic_metadata['list'] = ['foo', 'bar']
+        self.worker.metadata()['list'].append('baz')
 
-        self.assertEqual(
-            self.worker.dynamic_metadata['schroots'], ['buster', 'bullseye']
-        )
+        self.assertEqual(self.worker.dynamic_metadata['list'], ['foo', 'bar'])
 
     def test_set_dynamic_metadata(self) -> None:
         """Worker.set_dynamic_metadata sets the dynamic metadata."""
@@ -368,3 +552,10 @@ class WorkerTests(TestHelpersMixin, TestCase):
             self.worker.__str__(),
             f"Id: {self.worker.id} Name: {self.worker.name}",
         )
+
+    def test_get_absolute_url(self) -> None:
+        """Test the get_absolute_url method."""
+        self.assertEqual(
+            self.worker.get_absolute_url(),
+            f"/-/status/workers/{self.worker.name}/",
+        )
