@@ -30,7 +30,6 @@ from debusine.db.models import (
     CollectionItem,
     Group,
     User,
-    WorkRequest,
     WorkflowTemplate,
     Workspace,
     default_workspace,
@@ -439,8 +438,8 @@ class LookupSingleTests(LookupMixin, TestCase):
             template = WorkflowTemplate.objects.create(
                 name="test", workspace=default_workspace(), task_name="noop"
             )
-            workflow_root = WorkRequest.objects.create_workflow(
-                template=template, created_by=self.user, data={}
+            workflow_root = self.playground.create_workflow(
+                template, task_data={}
             )
             self.playground.create_collection("test", CollectionCategory.TEST)
 
@@ -1030,9 +1029,7 @@ class LookupMultipleTests(LookupMixin, TestCase):
         template = WorkflowTemplate.objects.create(
             name="test", workspace=default_workspace(), task_name="noop"
         )
-        workflow_root = WorkRequest.objects.create_workflow(
-            template=template, created_by=self.user, data={}
-        )
+        workflow_root = self.playground.create_workflow(template, task_data={})
 
         def do_lookup_filter(
             key: str, value: LookupSingle | LookupMultiple, **kwargs: Any
