@@ -15,7 +15,11 @@ from debusine.assets import KeyPurpose
 from debusine.client.models import LookupChildType
 from debusine.db.context import context
 from debusine.db.models import Artifact, TaskDatabase, WorkRequest
-from debusine.server.workflows import DebianPipelineWorkflow, workflow_utils
+from debusine.server.workflows import (
+    DebianPipelineWorkflow,
+    WorkflowValidationError,
+    workflow_utils,
+)
 from debusine.server.workflows.base import (
     WorkflowRunError,
     orchestrate_workflow,
@@ -46,7 +50,10 @@ class DebianPipelineWorkflowTests(TestCase):
         )
 
     def create_debian_pipeline_workflow(
-        self, *, extra_task_data: dict[str, Any]
+        self,
+        *,
+        extra_task_data: dict[str, Any] | None = None,
+        validate: bool = True,
     ) -> DebianPipelineWorkflow:
         """Create a debian pipeline workflow."""
         task_data = {
@@ -54,9 +61,10 @@ class DebianPipelineWorkflowTests(TestCase):
             "vendor": "debian",
             "codename": "bookworm",
         }
-        task_data.update(extra_task_data)
+        if extra_task_data is not None:
+            task_data.update(extra_task_data)
         wr = self.playground.create_workflow(
-            task_name="debian_pipeline", task_data=task_data
+            task_name="debian_pipeline", task_data=task_data, validate=validate
         )
         return DebianPipelineWorkflow(wr)
 
@@ -82,6 +90,30 @@ class DebianPipelineWorkflowTests(TestCase):
 
         return wr
 
+    def test_validate_input(self) -> None:
+        """validate_input passes a valid case."""
+        workflow = self.create_debian_pipeline_workflow()
+
+        workflow.validate_input()
+
+    def test_validate_input_bad_suite_collection(self) -> None:
+        """validate_input raises errors in looking up a suite."""
+        workflow = self.create_debian_pipeline_workflow(
+            extra_task_data={
+                "enable_reverse_dependencies_autopkgtest": True,
+                "reverse_dependencies_autopkgtest_suite": (
+                    "nonexistent@debian:suite"
+                ),
+            },
+            validate=False,
+        )
+
+        with self.assertRaisesRegex(
+            WorkflowValidationError,
+            "'nonexistent@debian:suite' does not exist or is hidden",
+        ):
+            workflow.validate_input()
+
     def test_compute_dynamic_data(self) -> None:
         """Dynamic data includes a parameter summary."""
         source_artifact = self.playground.create_source_artifact(
@@ -487,7 +519,7 @@ class DebianPipelineWorkflowTests(TestCase):
     ) -> None:
         """Populate raise WorkflowRunError: missing required field."""
         msg = (
-            'orchestrator failed: "make_signed_source_purpose" must be '
+            'Orchestrator failed: "make_signed_source_purpose" must be '
             'set when signing the source'
         )
         with self.assertRaisesRegex(WorkflowRunError, msg):
@@ -505,7 +537,7 @@ class DebianPipelineWorkflowTests(TestCase):
     ) -> None:
         """Populate raise WorkflowRunError: missing required field."""
         msg = (
-            'orchestrator failed: "make_signed_source_key" must be set '
+            'Orchestrator failed: "make_signed_source_key" must be set '
             'when signing the source'
         )
         with self.assertRaisesRegex(WorkflowRunError, msg):
@@ -879,5 +911,5 @@ class DebianPipelineWorkflowTests(TestCase):
 
     def test_get_label(self) -> None:
         """Test get_label."""
-        w = self.create_debian_pipeline_workflow(extra_task_data={})
+        w = self.create_debian_pipeline_workflow()
         self.assertEqual(w.get_label(), "run Debian pipeline")
