@@ -1,4 +1,4 @@
-# Copyright 2024 The Debusine Developers
+# Copyright © The Debusine Developers
 # See the AUTHORS file at the top-level directory of this distribution
 #
 # This file is part of Debusine. It is subject to the license terms
@@ -9,21 +9,31 @@
 
 """Unit tests for the package upload workflow."""
 from typing import Any
+from unittest.mock import call, patch
+
+try:
+    import pydantic.v1 as pydantic
+except ImportError:
+    import pydantic as pydantic  # type: ignore
 
 from debusine.artifacts.models import ArtifactCategory, CollectionCategory
+from debusine.db.context import context
 from debusine.db.models import (
     Artifact,
     CollectionItem,
+    TaskDatabase,
     WorkRequest,
     WorkflowTemplate,
     default_workspace,
 )
+from debusine.server.tasks.models import PackageUploadTarget
 from debusine.server.workflows import (
     PackageUploadWorkflow,
     WorkflowValidationError,
 )
 from debusine.server.workflows.base import orchestrate_workflow
-from debusine.tasks.models import LookupMultiple, TaskTypes
+from debusine.server.workflows.models import PackageUploadWorkflowData
+from debusine.tasks.models import BaseDynamicTaskData, LookupMultiple, TaskTypes
 from debusine.test.django import TestCase
 
 
@@ -34,6 +44,7 @@ class PackageUploadWorkflowTests(TestCase):
         self,
         *,
         extra_task_data: dict[str, Any],
+        validate: bool = True,
     ) -> PackageUploadWorkflow:
         """Create a package upload workflow."""
         task_data = {
@@ -42,14 +53,20 @@ class PackageUploadWorkflowTests(TestCase):
         }
         task_data.update(extra_task_data)
         wr = self.playground.create_workflow(
-            task_name="package_upload", task_data=task_data
+            task_name="package_upload", task_data=task_data, validate=validate
         )
         return PackageUploadWorkflow(wr)
 
-    def create_source_package(self) -> Artifact:
+    def create_source_package(self, *, name: str = "hello") -> Artifact:
         """Create a minimal `debian:source-package` artifact."""
         artifact, _ = self.playground.create_artifact(
-            category=ArtifactCategory.SOURCE_PACKAGE, data={}
+            category=ArtifactCategory.SOURCE_PACKAGE,
+            data={
+                "name": name,
+                "version": "1.0",
+                "type": "dpkg",
+                "dsc_fields": {},
+            },
         )
         return artifact
 
@@ -74,9 +91,11 @@ class PackageUploadWorkflowTests(TestCase):
 
         return source, binary
 
-    def test_validate_input_source_binary_artifacts_not_set(self):
+    def test_validate_input_source_binary_artifacts_not_set(self) -> None:
         """Invalid: source_artifact and binaries_artifacts not set."""
-        w = self.create_package_upload_workflow(extra_task_data={})
+        w = self.create_package_upload_workflow(
+            extra_task_data={}, validate=False
+        )
 
         with self.assertRaisesRegex(
             WorkflowValidationError,
@@ -84,28 +103,16 @@ class PackageUploadWorkflowTests(TestCase):
         ):
             w.validate_input()
 
-    def test_validate_input_mergeuploads_set_vendor_codename_not_set(self):
-        """Raise WorkflowValidationError: vendor and codename required."""
-        w = self.create_package_upload_workflow(
-            extra_task_data={"merge_uploads": True, "binary_artifacts": [1]}
-        )
-
-        with self.assertRaisesRegex(
-            WorkflowValidationError,
-            '"vendor" and "codename" are required '
-            'when "merge_uploads" is set',
-        ):
-            w.validate_input()
-
-    def test_validate_input_source_package_vendor_codename_not_set(self):
+    @context.disable_permission_checks()
+    def test_validate_input_source_package_vendor_codename_not_set(
+        self,
+    ) -> None:
         """Raise WorkflowValidationError: vendor and codename required."""
         source = self.create_source_package()
 
         w = self.create_package_upload_workflow(
-            extra_task_data={
-                "merge_uploads": True,
-                "source_artifact": source.id,
-            }
+            extra_task_data={"source_artifact": source.id},
+            validate=False,
         )
 
         with self.assertRaisesRegex(
@@ -117,7 +124,7 @@ class PackageUploadWorkflowTests(TestCase):
 
     def test_create_orchestrator(self) -> None:
         """A PackageUploadWorkflow can be instantiated."""
-        source_artifact = 2
+        source_artifact = self.create_source_package().id
         binary_artifacts = ["internal@collections/name:build-arm64"]
         target_distribution = "debian:bookworm"
 
@@ -126,6 +133,8 @@ class PackageUploadWorkflowTests(TestCase):
                 "source_artifact": source_artifact,
                 "binary_artifacts": binary_artifacts,
                 "target_distribution": target_distribution,
+                "vendor": "debian",
+                "codename": "bookworm",
             }
         )
 
@@ -144,10 +153,9 @@ class PackageUploadWorkflowTests(TestCase):
             task_name="package_upload",
         )
 
-        wr = WorkRequest.objects.create_workflow(
-            template=template,
-            data={**{"target": "sftp://upload.example.org/queue"}, **data},
-            created_by=self.playground.get_default_user(),
+        wr = self.playground.create_workflow(
+            template,
+            task_data={**{"target": "sftp://upload.example.org/queue"}, **data},
         )
 
         wr.mark_running()
@@ -163,18 +171,24 @@ class PackageUploadWorkflowTests(TestCase):
         -makesourcepackageupload
         -package_upload
         """
-        source = self.create_source_package()
+        with context.disable_permission_checks():
+            source = self.create_source_package()
 
-        wr = self.orchestrate(
-            data={
-                "source_artifact": f"{source.id}@artifacts",
-                "target_distribution": "debian:bookworm",
-                "since_version": "1.0",
-                "require_signature": False,
-                "vendor": "debian",
-                "codename": "bullseye",
-            }
-        )
+        with patch.object(
+            PackageUploadWorkflow,
+            "requires_artifact",
+            wraps=PackageUploadWorkflow.requires_artifact,
+        ) as requires_artifact:
+            wr = self.orchestrate(
+                data={
+                    "source_artifact": source.id,
+                    "target_distribution": "debian:bookworm",
+                    "since_version": "1.0",
+                    "require_signature": False,
+                    "vendor": "debian",
+                    "codename": "bullseye",
+                }
+            )
 
         self.assertEqual(wr.children.count(), 2)
 
@@ -192,7 +206,7 @@ class PackageUploadWorkflowTests(TestCase):
             make_source.task_data,
             {
                 "environment": "debian/match:codename=bullseye",
-                "input": {"source_artifact": f"{source.id}@artifacts"},
+                "input": {"source_artifact": source.id},
                 "since_version": "1.0",
                 "target_distribution": "debian:bookworm",
             },
@@ -218,6 +232,13 @@ class PackageUploadWorkflowTests(TestCase):
             },
         )
 
+        # Assert that self.requires_artifact() was called.
+        # It cannot test the dependencies because source_artifact was
+        # a real artifact and not a promise (so no dependencies were added)
+        self.assertIn(
+            call(make_source, source.id), requires_artifact.mock_calls
+        )
+
         [package_upload] = wr.children.filter(task_name="packageupload")
         self.assertEqual(package_upload.task_type, TaskTypes.SERVER)
         self.assertEqual(
@@ -240,6 +261,7 @@ class PackageUploadWorkflowTests(TestCase):
                     "upload": "internal@collections/name:package-upload-source"
                 },
                 "target": "sftp://upload.example.org/queue",
+                "delayed_days": None,
             },
         )
 
@@ -250,15 +272,13 @@ class PackageUploadWorkflowTests(TestCase):
             ),
         )
 
-    def test_populate_source_is_package_upload(self):
+    def test_populate_source_is_package_upload(self) -> None:
         """Category source package is UPLOAD: skip makesourcepackageupload."""
-        source = self.create_source_package()
-        source.category = ArtifactCategory.UPLOAD
-        source.save()
+        source = self.playground.create_upload_artifacts().upload
 
         wr = self.orchestrate(
             data={
-                "source_artifact": f"{source.id}@artifacts",
+                "source_artifact": source.id,
                 "target_distribution": "debian:bookworm",
                 "since_version": "1.0",
                 "require_signature": False,
@@ -271,21 +291,23 @@ class PackageUploadWorkflowTests(TestCase):
         self.assertEqual(
             package_upload.workflow_data_json,
             {
-                "display_name": f"Package upload {source.id}@artifacts",
-                "step": f"package-upload-{source.id}@artifacts",
+                "display_name": f"Package upload {source.id}",
+                "step": f"package-upload-{source.id}",
             },
         )
         self.assertEqual(
             package_upload.task_data,
             {
-                "input": {"upload": f"{source.id}@artifacts"},
+                "input": {"upload": source.id},
                 "target": "sftp://upload.example.org/queue",
+                "delayed_days": None,
             },
         )
 
     def test_populate_package_uploads_binaries_only(self) -> None:
         """The workflow create children package uploads: binaries only."""
-        (_, binary) = self.create_source_binary_artifacts()
+        with context.disable_permission_checks():
+            (_, binary) = self.create_source_binary_artifacts()
 
         wr = self.orchestrate(
             data={
@@ -302,26 +324,27 @@ class PackageUploadWorkflowTests(TestCase):
         # Other properties of the created package is tested on different
         # unit tests
 
-    def test_populate_merge_uploads(self):
+    def test_populate_merge_uploads(self) -> None:
         """Test mergeuploads task is created (merge_uploads=True)."""
-        (source, binary_1) = self.create_source_binary_artifacts()
+        with context.disable_permission_checks():
+            (source, binary_1) = self.create_source_binary_artifacts()
 
-        binary_2 = self.playground.create_minimal_binary_packages_artifact(
-            "hello", "1.0-1", "1.0-1", "amd64"
-        )
+            binary_2 = self.playground.create_minimal_binary_packages_artifact(
+                "hello", "1.0-1", "1.0-1", "amd64"
+            )
 
-        CollectionItem.objects.create_from_artifact(
-            binary_2,
-            parent_collection=binary_1.parent_collections.first(),
-            name="binary-2",
-            created_by_user=self.playground.get_default_user(),
-            data={},
-            created_by_workflow=None,
-        )
+            CollectionItem.objects.create_from_artifact(
+                binary_2,
+                parent_collection=binary_1.parent_collections.earliest("id"),
+                name="binary-2",
+                created_by_user=self.playground.get_default_user(),
+                data={},
+                created_by_workflow=None,
+            )
 
         wr = self.orchestrate(
             data={
-                "source_artifact": f"{source.id}@artifacts",
+                "source_artifact": source.id,
                 "binary_artifacts": [binary_1.id, binary_2.id],
                 "require_signature": False,
                 "target_distribution": "debian/match:codename=bookworm",
@@ -340,7 +363,6 @@ class PackageUploadWorkflowTests(TestCase):
         self.assertEqual(
             merge_uploads.task_data,
             {
-                "environment": "debian/match:codename=bullseye",
                 "input": {
                     "uploads": [
                         "internal@collections/name:package-upload-source",
@@ -353,7 +375,9 @@ class PackageUploadWorkflowTests(TestCase):
 
         self.assertQuerySetEqual(
             merge_uploads.dependencies.all(),
-            WorkRequest.objects.filter(task_name="makesourcepackageupload"),
+            list(
+                WorkRequest.objects.filter(task_name="makesourcepackageupload")
+            ),
         )
 
         self.assertEqual(
@@ -367,30 +391,35 @@ class PackageUploadWorkflowTests(TestCase):
                     "upload": "internal@collections/name:package-upload-merged"
                 },
                 "target": "sftp://upload.example.org/queue",
+                "delayed_days": None,
             },
         )
 
-    def test_upload_debsign(self):
+    def test_upload_debsign(self) -> None:
         """Upload signed artifacts."""
-        source, binary = self.create_source_binary_artifacts()
-        key, _ = self.create_artifact(category=ArtifactCategory.SIGNING_KEY)
+        with context.disable_permission_checks():
+            source, binary = self.create_source_binary_artifacts()
 
         wr = self.orchestrate(
             data={
-                "source_artifact": f"{source.id}@artifacts",
+                "source_artifact": source.id,
                 "binary_artifacts": [binary.id],
                 "require_signature": False,
                 "target_distribution": "debian/match:codename=bookworm",
                 "merge_uploads": True,
-                "key": f"{key.id}@artifacts",
+                "key": "ABC123",
+                "vendor": "debian",
+                "codename": "bookworm",
             },
         )
 
-        [signer] = wr.children.filter(task_name="debsign")
+        [signer] = wr.children.filter(
+            task_type=TaskTypes.SIGNING, task_name="debsign"
+        )
         self.assertEqual(
             signer.task_data,
             {
-                "key": f"{key.id}@artifacts",
+                "key": "ABC123",
                 "unsigned": "internal@collections/name:package-upload-merged",
             },
         )
@@ -403,6 +432,14 @@ class PackageUploadWorkflowTests(TestCase):
             },
         )
 
+        self.assertQuerySetEqual(
+            signer.dependencies.all(),
+            list(
+                WorkRequest.objects.filter(
+                    task_type=TaskTypes.WORKER, task_name="mergeuploads"
+                )
+            ),
+        )
         signed_name = (
             "package-upload-signed-"
             "internal_collections_name_package-upload-merged"
@@ -431,6 +468,7 @@ class PackageUploadWorkflowTests(TestCase):
             {
                 "input": {"upload": f"internal@collections/name:{signed_name}"},
                 "target": "sftp://upload.example.org/queue",
+                "delayed_days": None,
             },
         )
 
@@ -446,21 +484,26 @@ class PackageUploadWorkflowTests(TestCase):
             },
         )
 
-    def test_upload_externaldebsign(self):
+    def test_upload_externaldebsign(self) -> None:
         """Upload external signed artifacts."""
-        source, binary = self.create_source_binary_artifacts()
+        with context.disable_permission_checks():
+            source, binary = self.create_source_binary_artifacts()
 
         wr = self.orchestrate(
             data={
-                "source_artifact": f"{source.id}@artifacts",
+                "source_artifact": source.id,
                 "binary_artifacts": [binary.id],
                 "target_distribution": "debian/match:codename=bookworm",
                 "merge_uploads": True,
                 "require_signature": True,
+                "vendor": "debian",
+                "codename": "bookworm",
             },
         )
 
-        [signer] = wr.children.filter(task_name="externaldebsign")
+        [signer] = wr.children.filter(
+            task_type=TaskTypes.WAIT, task_name="externaldebsign"
+        )
         self.assertEqual(
             signer.task_data,
             {"unsigned": "internal@collections/name:package-upload-merged"},
@@ -473,9 +516,18 @@ class PackageUploadWorkflowTests(TestCase):
                     "Wait for signature on " f"upload for {identifier}"
                 ),
                 "step": f"external-debsign-{identifier}",
+                "needs_input": True,
             },
         )
 
+        self.assertQuerySetEqual(
+            signer.dependencies.all(),
+            list(
+                WorkRequest.objects.filter(
+                    task_type=TaskTypes.WORKER, task_name="mergeuploads"
+                )
+            ),
+        )
         signed_name = (
             "package-upload-signed-"
             "internal_collections_name_package-upload-merged"
@@ -504,17 +556,49 @@ class PackageUploadWorkflowTests(TestCase):
             {
                 "input": {"upload": f"internal@collections/name:{signed_name}"},
                 "target": "sftp://upload.example.org/queue",
+                "delayed_days": None,
             },
         )
 
-    def test_orchestrate_idempotent(self):
-        """Calling orchestrate twice does not create new work requests."""
-        source, binary = self.create_source_binary_artifacts()
+    def test_upload_delayed(self) -> None:
+        """Upload to a delayed queue."""
+        with context.disable_permission_checks():
+            source = self.create_source_package()
 
         wr = self.orchestrate(
             data={
-                "source_artifact": f"{source.id}@artifacts",
+                "source_artifact": source.id,
+                "delayed_days": 3,
+                "vendor": "debian",
+                "codename": "bookworm",
+            }
+        )
+
+        signed_name = (
+            "package-upload-signed-"
+            "internal_collections_name_package-upload-source"
+        )
+        uploader = wr.children.filter(task_name="packageupload").get()
+        self.assertEqual(
+            uploader.task_data,
+            {
+                "input": {"upload": f"internal@collections/name:{signed_name}"},
+                "target": "sftp://upload.example.org/queue",
+                "delayed_days": 3,
+            },
+        )
+
+    def test_orchestrate_idempotent(self) -> None:
+        """Calling orchestrate twice does not create new work requests."""
+        with context.disable_permission_checks():
+            source, binary = self.create_source_binary_artifacts()
+
+        wr = self.orchestrate(
+            data={
+                "source_artifact": source.id,
                 "binary_artifacts": [binary.id],
+                "vendor": "debian",
+                "codename": "bookworm",
             },
         )
 
@@ -524,12 +608,65 @@ class PackageUploadWorkflowTests(TestCase):
 
         self.assertQuerySetEqual(wr.children.all(), children, ordered=False)
 
-    def test_get_label(self):
+    def test_compute_dynamic_data_source_artifact(self) -> None:
+        source_artifact = self.create_source_package(name="hello")
+
+        wr = self.playground.create_workflow(
+            task_name="package_upload",
+            task_data=PackageUploadWorkflowData(
+                source_artifact=source_artifact.id,
+                target=pydantic.parse_obj_as(
+                    PackageUploadTarget,
+                    "sftp://example.org",
+                ),
+                vendor="debian",
+                codename="trixie",
+            ),
+        )
+        workflow = PackageUploadWorkflow(wr)
+
+        self.assertEqual(
+            workflow.compute_dynamic_data(TaskDatabase(wr)),
+            BaseDynamicTaskData(subject="hello"),
+        )
+
+    def test_compute_dynamic_data_binary_artifacts(self) -> None:
+        binary_artifact = (
+            self.playground.create_minimal_binary_package_artifact(
+                srcpkg_name="hello"
+            )
+        )
+
+        wr = self.playground.create_workflow(
+            task_name="package_upload",
+            task_data=PackageUploadWorkflowData(
+                binary_artifacts=LookupMultiple.parse_obj([binary_artifact.id]),
+                source_artifact=None,
+                target=pydantic.parse_obj_as(
+                    PackageUploadTarget,
+                    "sftp://example.org",
+                ),
+                vendor="debian",
+                codename="trixie",
+            ),
+        )
+        workflow = PackageUploadWorkflow(wr)
+
+        self.assertEqual(
+            workflow.compute_dynamic_data(TaskDatabase(wr)),
+            BaseDynamicTaskData(subject="hello"),
+        )
+
+    @context.disable_permission_checks()
+    def test_get_label(self) -> None:
         """Test get_label()."""
+        source = self.create_source_package()
         w = self.create_package_upload_workflow(
             extra_task_data={
-                "source_artifact": 2,
+                "source_artifact": source.id,
                 "target_distribution": "debian:bookworm",
+                "vendor": "debian",
+                "codename": "bookworm",
             }
         )
         self.assertEqual(w.get_label(), "run package uploads")
