@@ -12,12 +12,13 @@
 import io
 import re
 from collections.abc import Callable
-from datetime import datetime
+from datetime import datetime, timezone
 from typing import Any, TypeVar, Union, cast
 from unittest import mock
 
 import lxml.etree
 import lxml.objectify
+import yaml
 from django.contrib.auth.models import AnonymousUser
 from django.http import HttpRequest
 from django.test import RequestFactory
@@ -98,6 +99,24 @@ class ViewTestMixin(BaseDjangoTestCase):
             errors.append(f"{error.line}:{error.type_name}:{error.message}")
         return errors
 
+    def assertResponseHTML(
+        self,
+        response: TestResponseType,
+        status_code: int = status.HTTP_200_OK,
+        content_type: str = "text/html; charset=utf-8",
+    ) -> lxml.objectify.ObjectifiedElement:
+        """
+        Check that the response is valid HTML.
+
+        :param response: the Django response to check
+        :param status_code: expected status code
+        :param content_type: expected content type
+        :returns: the parsed HTML contents
+        """
+        self.assertEqual(response.status_code, status_code)
+        self.assertEqual(response.headers["Content-Type"], content_type)
+        return self.assertHTMLValid(response)
+
     def assertHTMLValid(
         self, response: "Union[TestResponseType, str]"
     ) -> lxml.objectify.ObjectifiedElement:
@@ -119,10 +138,10 @@ class ViewTestMixin(BaseDjangoTestCase):
             lxml.objectify.ObjectifyElementClassLookup()
         )
         with io.BytesIO(content) as fd:
-            root = lxml.etree.parse(fd, parser)
+            tree = lxml.etree.parse(fd, parser)
         errors = self._filter_parser_error_log(parser.error_log)
         self.assertEqual(errors, [])
-        return cast(lxml.objectify.ObjectifiedElement, root)
+        return cast(lxml.objectify.ObjectifiedElement, tree.getroot())
 
     def assertHasElement(
         self, tree: lxml.objectify.ObjectifiedElement, xpath: str
@@ -199,6 +218,38 @@ class ViewTestMixin(BaseDjangoTestCase):
         text = " ".join(text.strip().split())
         self.assertEqual(sample, text)
 
+    def assertDatetimeContentEqual(
+        self, node: lxml.objectify.ObjectifiedElement, dt: datetime
+    ) -> None:
+        """
+        Ensure the row contains a datetime matching the given one.
+
+        Text surrounding the datetime value is ignored.
+        """
+        text = self.get_node_text_normalized(node)
+        mo = re.search(r"\d+-\d+-\d+ \d+:\d+", text)
+        if not mo:
+            self.fail(
+                f"text content {text!r} does not contain"
+                " a YYYY-MM-DD HH:MM datetime"
+            )
+        actual = datetime.strptime(mo.group(), "%Y-%m-%d %H:%M").replace(
+            tzinfo=timezone.utc
+        )
+        expected = dt.replace(second=0, microsecond=0)
+        self.assertEqual(actual, expected)
+
+    def assertYAMLContentEqual(
+        self, node: lxml.objectify.ObjectifiedElement, data: Any
+    ) -> None:
+        """Ensure that the node contents match yaml-formatted data."""
+        encoded = "".join(node.itertext())
+        try:
+            actual = yaml.safe_load(encoded)
+        except yaml.parser.ParserError as e:
+            self.fail(f"node text {encoded!r} does not parse as YAML: {e}")
+        self.assertEqual(actual, data)
+
     def assertWorkRequestRow(
         self, tr: lxml.objectify.ObjectifiedElement, work_request: WorkRequest
     ) -> None:
